# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-02 17:11 (EST)
# Function: Monitor Element class
#
# $Id: MonEl.pm,v 1.39 2002/10/20 22:32:14 jaw Exp jaw $

package MonEl;
use Stats;
use Override;
use Web;
#use Tkt;

@all = ();
%byname = ();

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [ ],
    methods => {},
    fields  =>  {
	i_am_top => {},
	transient=> {
	    descr => 'object is temporary',
	},
	confck   => {},
	name     => {
	    descr=> 'relative name of this object',
	},
	type     => {
	    descr => 'type of object',
	},
	uname    => {
	    # will be same as name, except for services
	    descr => 'relative unique name for this object',
	    attrs => ['config'],
	    # eg. Group "Foo" {
	    #		Service HTTP
	    #		Service HTTP {
	    #			port: 8080
	    #  			uname: HTTP-2
	    #		}
	}, 
	unique   => {
	    descr => 'full unique name of this object',
	},
	parents  => {
	    descr => 'list of objects parents',
	},
	children => {
	    descr => 'list of objects children',
	},
	override => {
	    descr => 'override data',
	},
	anc_in_ov => {
	    descr => 'an ancestor is in override',
	},
	alarm    => {
	    descr => 'is the siren currently ringing',
	    attrs => ['bool'],
	},
	
	definedinfile => {
	    descr => 'file the object was defined in',
	},
	definedonline => {
	    descr => 'line number the object was defined on',
	},
	
	sort => {
	    descr => 'sort the entries on web page',
	    default => 'yes', # except for Top
	    attrs => ['config', 'bool'],
	},
	autogenerated => {
	    descr => 'was this object generated automagically',
	    # to be used by config file editing utilities
	    # or built-in web-based configurator (RSN)
	    default => 'no',
	    attrs => ['config', 'bool', 'inherit'],
	},
	overridable => {
	    descr => 'can this object be overridden', # but see also acl_override
	    attrs => ['config', 'bool'],
	    default => 'yes',
	},
	note => {
	    descr => 'informational text displayed on webpage',
	    attrs => ['config'],
	},
	info => {
	    descr => 'informational text displayed on webpage',
	    attrs => ['config'],
	},
	comment => {
	    descr => 'informational text displayed on webpage',
	    attrs => ['config'],
	},
	details => {
	    # expected to be a <A HREF="link">label</A>
	    descr => 'informational text displayed on webpage - usually a link to more details',
	    attrs => ['config'],
	},
	debug => {
	    descr => 'send all sorts of gibberish to the log',
	    default => 'no',
	    attrs => ['config', 'bool'],
	},
	flags => {
	    descr => 'various config data, summarized and displayed on webpage',
	    attrs => ['internal'],
	},
	passive => {
	    descr => 'marks object as passive - sets nostatus=yes, siren=no, sendnotify=no',
	    attrs => ['config', 'inherit', 'bool'],
	    default => 'no',
	},
	
	# transition
	siren => {
	    descr => 'should alarm ring',
	    attrs => ['config', 'inherit', 'bool'],
	    default => 'on',
	},
	nostatus => {
	    descr => 'should the status be ignored',
	    attrs => ['config', 'inherit', 'bool'],
	    default => 'no',
	},
	# Change the gravitational constant of the universe!
	#   -- Q, Star Trek TNG
	gravity => {
	    descr => 'governs group state transitions: down => group is down if any child is down, up => group is up if any child is up',
	    attrs => ['config'],
	    default => 'down',
	    vals  => ['up', 'down'],
	    # down => object is down if any child is down, (even though some are up)
	    # up   => object is up if any child is up, (even though some are down)
	    #
	    # eg. set gravity=up if you want to notified only if ALL nameservers are down...
	},
	countstop => {
	    descr => 'do not recurse and count children in count mode (the numbers on Top webpage)',
	    attrs => ['config', 'bool'],
	    default => 'no',
	},
	
	status => {
	    descr => 'current real status',
	    vals  => ['up', 'down'], # degraded, attention, bouncing
	},
	ovstatus => {
	    descr => 'current status',
	    vals  => ['up', 'down', 'override'],
	},
	prevstatus   => {},
	prevovstatus => {},
	
	transtime => {
	    descr => 'time of last transition',
	},
	sirentime => {},
	
	# notification
      notify::messageup => {
	  descr => 'message to transmit for UP notifications',
	  attrs => ['config'],
      },
      notify::messagedn => {
	  descr => 'message to transmit for DOWN notifications',
	  # "Server Room is too Hot" may be better than "1.2.3.4:.1.2.4.6.8.35.17.24.7.1 is down"
	  # NB: messages can be emtpy, in which case nothing will be sent
	  attrs => ['config'],
      },
      notify::notify => {
	  descr => 'list of addresses to notify: method:address method:address ...',
	  # currently, method is 'email' or 'qpage'
	  attrs => ['config', 'inherit'],
      },
      notify::renotify => {
	  descr => 're-send time period in seconds',
	  attrs => ['config', 'inherit'],
	  default => 300,
      },
      notify::autoack => {
	  descr => 'do notifications not require ack',
	  attrs => ['config', 'inherit', 'bool'],
	  default => 'no',
      },
      notify::escalate => {
	  descr => 'list of times (in minutes) and addresses for notification escalation: time method:address ...; time method:address ...',
	  attrs => ['config', 'inherit'],
	  # eg. escalate: 30 qpage:manager qpage:noc; 60 qpage:ceo
      },
      notify::sendnotify => {
	  descr => 'should notifications be sent for this object',
	  attrs => ['config', 'bool'],
	  # default varies by object, Services=>yes, Group=>no
      },
      notify::sendnotify2 => {
	  descr => 'should notifications be sent for this Service',
	  attrs => ['config', 'bool', 'inherit'],
	  default => 'yes',
      },
	
      notify::priority => {
	  # not currently used
	  descr => 'priority level of notifications',
	  attrs => ['config', 'inherit'],
      },
	
      notify::list => {
	  descr => 'list of all outstanding notifications',
      },
	
	# web page
      web::transtime => {
	  descr => 'last time the object changed',
      },
      web::bldtime   => {
	  descr => 'time the web page was last generated',
      },
      web::bldtimetop=> {
	  descr => 'time the web page (top version) was last generated',
      },

      # there is a reason there is both a header and header_all (likewise footer)
      # but I forget why....
      web::header => {
	  descr => 'stuff to put at the top of the webpage',
	  attrs => ['config'],
      },
      web::footer => {
	  descr => 'stuff to put at the bottom of the webpage',
	  attrs => ['config'],
      },
      web::header_all => {
	  descr => 'stuff to put at the top of the webpage',
	  attrs => ['config', 'inherit'],
      },
      web::footer_all => {
	  descr => 'stuff to put at the bottom of the webpage',
	  attrs => ['config', 'inherit'],
      },
      web::showstats => {
	  descr => 'show stats on the web page',
	  attrs => ['config', 'bool'],
	  default => 'yes',
      },
      web::bkgimage => {
	  descr => 'web page background image',
	  attrs => ['config', 'inherit'],
      },
      web::refresh => {
	  descr => 'web page refresh period',
	  attrs => ['config', 'inherit'],
	  default => 60,
      },
      web::icon => {
	  descr => '16x16 icon shown in some browsers',
	  attrs => ['config', 'inherit'],
      },
      web::icon_up => {
	  descr => '16x16 icon shown in some browsers - when object is up',
	  attrs => ['config', 'inherit'],
      },
      web::icon_down => {
	  descr => '16x16 icon shown in some browsers - when object is down',
	  attrs => ['config', 'inherit'],
      },
      web::nospkr_icon => {
	  descr => 'icon',
	  attrs => ['config', 'inherit'],
      },
	
	label_left => {
	    descr => 'label to use on the left side of the webpage',
	    attrs => ['config'],
	},
	label_right => {
	    descr => 'label to use on the right side of the webpage',
	    attrs => ['config'],
	},
	label_right_maybe => {},
	label_left_maybe  => {},
	label => {
	    descr => 'label to use on right and/or left side of the webpage',
	    attrs => ['config'],
	},

	#### ACLs
	acl_mode => {
	    descr => 'mode for access control list configuration',
	    attrs => ['config', 'inherit'],
	    vals  => ['simple', 'extended'],
	    default => 'simple',
	},

	# acl_mode = simple:
	acl_user => {
	    descr => 'simple mode access control list - read-only web page access',
	    attrs => ['config', 'acl'],
	    default => 'user staff root',
	},
	acl_staff => {
	    descr => 'simple mode access control list - some read/write web page access',
	    attrs => ['config', 'acl'],
	    default => 'staff root',
	},
	acl_root => {
	    descr => 'simple mode access control list - full web page access',
	    attrs => ['config', 'acl'],
	    default => 'root',
	},

	# acl_mode = extended:
	acl_page => {
	    descr => 'extended mode access control list - for web page access',
	    attrs => ['config', 'acl'],
	},
	acl_about => {
	    descr => 'extended mode access control list - for debugging page access',
	    attrs => ['config', 'acl'],
	},
	acl_getconf => {
	    descr => 'extended mode access control list - display config page access',
	    attrs => ['config', 'acl'],
	},
	acl_flush => {
	    descr => 'extended mode access control list - flush page access',
	    attrs => ['config', 'acl'],
	},
	acl_override => {
	    descr => 'extended mode access control list - set/clear override access',
	    attrs => ['config', 'acl'],
	},
	acl_annotate => {
	    descr => 'extended mode access control list - add/remove annotation access',
	    attrs => ['config', 'acl'],
	},
	acl_logfile => {
	    descr => 'extended mode access control list - logfile access',
	    attrs => ['config', 'acl'],
	},
	acl_ntfylist => {
	    descr => 'extended mode access control list - notification list access',
	    attrs => ['config', 'acl'],
	},
	acl_ntfyack => {
	    descr => 'extended mode access control list - ack notifications access',
	    attrs => ['config', 'acl'],
	},
	acl_ntfyackall => {
	    descr => 'extended mode access control list - ack all notifications access',
	    attrs => ['config', 'acl'],
	},
	acl_ntfydetail => {
	    descr => 'extended mode access control list - notification details page access',
	    attrs => ['config', 'acl'],
	},

	
	# stats
	logsize => {
	    descr => 'Maximum number of log entries to keep',
	    default => 200,
	    attrs => ['config', 'inherit'],
	},
	nostats => {
	    descr => 'do not record any statistics',
	    attrs => ['config', 'bool'],
	    default => 'no',
	},
      stats::log => {
	  descr => 'object event log - array of [Time, status, ovstatus, tag, msg]',
      },
      stats::lasttime => {
	  descr => 'time stats were last updated',
      },
      stats::status   => {
	  descr => 'status at time of last stats update',
      },
      stats::daily    => {
	  descr => 'daily statistics',
      },
      stats::monthly  => {
	  descr => 'monthly statistics',
      },
      stats::yearly   => {
	  descr => 'yearly statistics',
      },

################################################################
# experimental, undocumented features - for author's use only
################################################################
	_no_addtnl => {},
	_save_less => {},
	_ctl_debug => {},
	_hide_expr => {},
	_dont_ntfy => {},

      # for ticket system interface
      watch::watching => {},
      watch::watched  => {},

    },
};

sub unique {
    my $me = shift;
    my( $x, $n, $nn );
    
    return $me->{unique} if $me->{unique};

    if( $me->{parents} && $me->{parents}[0] ){
	$n = $me->{parents}[0]->unique();
	$n .= ':';
    }
    $nn = $me->{uname} || $me->{name};
    $nn =~ s/\s+/_/g;
    $nn =~ tr/\000-\037//d;
    $n .= $nn;
    $me->{unique} = $n;
    $n
}

sub filename {
    my $me = shift;
    my $n = $me->unique();

    return ::xxx_encode( $n );
}

sub url {
    my $me = shift;
    my @param = @_;
    my $n = $me->filename();

    return "__BASEURL__?object=$n" . join('&', '', @param);
}

sub init {
    my $me = shift;
    my $cf = shift;
    my( $u, $k, $v );

    $me->init_from_config( $doc, '' );      # this may change uname
    $me->init_from_config( $doc, 'web' );
    $me->init_from_config( $doc, 'stats' );
    $me->init_from_config( $doc, 'notify' );

    $me->init_acls();
    
    $me->{uname} ||= $me->{name};
    $u = $me->unique();
    if( $byname{$u} ){
	return $cf->error( "Duplicate object '$u'" );
    }

    $byname{$u} = $me;
    push @all, $me;

    $me->{label_left}  = $me->{label_left}  || $me->{label} || $me->{label_left_maybe};
    $me->{label_right} = $me->{label_right} || $me->{label} || $me->{label_right_maybe};
    
    $me->{status} = $me->{ovstatus} = 'up';
    $me->{alarm}  = 0;
    $me->{transtime} = $^T;
    $me->{sirentime} = $^T;

    $me->{notify}{messageup} ||= $me->unique() . ' is UP';
    $me->{notify}{messagedn} ||= $me->unique() . ' is DOWN';

    if( $me->{passive} ){
	$me->{flags} .= " *passive*";
	$me->{siren}    = 0;
	$me->{nostatus} = 1;
	$me->{notify}{sendnotify} = 0;
	# QQQ - set an override ?
    }
    
    $me->{flags} .= " status-ignored" if $me->{nostatus};
    $me->{flags} .= " gravity-up"     if $me->{gravity} eq 'up';
    $me->{flags} .= " siren-quiet"    unless $me->{siren};
    $me->{flags} .= " nostats"        if $me->{nostats};

    $cf->check_typos( $me ) if $cf;
    Doc::check( $me )       if $::opt_T;
    $me->stats_load()       unless $::opt_t;
    1;
}

END {
    my @a = @all;	# because recycle modifies @all
    
    return if $::opt_t;
    foreach my $x (@a){
	$x->stats_update_and_maybe_roll();
	$x->stats_save();
	$x->recycle();
    }
}

# initialize from config data according to method described in $doc
# hmmm, perhaps $doc is a poor choice of names....
sub init_from_config {
    my $me   = shift;
    my $doc  = shift;
    my $base = shift;

    foreach my $k (keys %{$doc->{fields}}){
	# ignore fields that are not marked configurable
	next unless has_attr($k, $doc, 'config');
	$me->init_field_from_config($doc, $base, $k);
    }
}

sub init_field_from_config {
    my $me   = shift;
    my $doc  = shift;
    my $base = shift;
    my $fld  = shift;
    my( $v, $kk );

    $kk = $fld;
    $kk =~ s/^($base)::// if $base;
    return if( $kk =~ /::/ );
    return if( $base && $fld !~ /::/ );

    # for detecting config file typos
    $me->{confck}{$kk} = 1;
    
    # does this feild inherit from parent if not set?
    if( has_attr($fld, $doc, 'inherit') ){
	$v = $me->inherit( $kk );
    }elsif( has_attr($fld, $doc, 'acl') ){
	$v = $me->build_acl($kk);
    }else{
	$v = $me->{config}{$kk};
    }

    # if not set, is there a default
    $v = $doc->{fields}{$fld}{default} unless defined($v);
    
    if( has_attr($fld, $doc, 'bool') ){
	$v = ::ckbool($v) if defined($v);
    }

    # install value
    if( $base ){
	$me->{$base}{$kk} = $v if defined($v);
    }else{
	$me->{$kk} = $v if defined($v);
    }

}

sub recycle {
    my $me = shift;

    # QQQ - remove links me->{parents}{children} ???
    delete $byname{ $me->unique() };
    @all = grep { $_->unique() ne $me->unique() } @all;
}

sub debug {
    my $me  = shift;
    my $msg = shift;
    my( $f );
    
    return unless $me->{debug} || $Conf::Config{debug};
    $f = fileno($me->{fd});
    $me->loggit( msg => "[$f] $msg",
		 tag => 'DEBUG' );
}

# { tag msg lpf slp }
# lpf: bool, log to logfile?
# slp: bool, add to object log?
# lss: log status
# also logs to syslog, consoles, STDERR (-f)
sub loggit {
    my $me = shift;
    my %p  = @_;
    my( $m, $msg );

    $msg = $p{msg} || $me->{ovstatus};
    
    if( $p{slp} ){
	push @{$me->{stats}{log}}, [ $^T, $me->{status}, $me->{ovstatus}, $p{tag}, $msg ];
	# and trim...
	if( @{$me->{stats}{log}} > $me->{logsize} ){
	    my( $l );
	    
	    $l = @{$me->{stats}{log}};
	    splice @{$me->{stats}{log}}, 0, $l - $me->{logsize}, ();
	}
    }
    
    $m = $p{tag} . ($p{tag} ? ' - ' : '') .
	( $p{lss} ? "$me->{ovstatus} - " : '') .
	$me->unique() . " - $msg";
    
    ::loggit( $m, $p{lpf} );
}

# And he said unto him, I am the LORD that brought thee out of
# Ur of the Chaldees, to give thee this land to inherit it.
#     -- genesis 15:7
sub inherit {
    my $me = shift;
    my $field = shift;

    $me->{confck}{$field} = 1;
    return $me->{config}{$field} if defined $me->{config}{$field};
    # special case Host "hostname" {...}
    return $me->{name} if( $me->{type} eq 'Host' && $field eq 'hostname' );
    return $me->{parents}[0]->inherit($field) if $me->{parents}[0];
    return $Conf::Config{$field} if defined $Conf::Config{$field};
    undef;
}

# build cumulative acl
sub build_acl {
    my $me = shift;
    my $field = shift;
    my( $pv, $mv, $v, %g );

    # parent values, then mine
    if( $me->{parents}[0] ){
	$pv = $me->{parents}[0]->build_acl($field) || '';
	# yes, this could be done smarter, ...
    }else{
	$pv = $Conf::Config{$field} || '';
    }
    $mv = $me->{config}{$field} || '';
    $v = "$pv $mv";

    # process directives, etc
    foreach my $g (split /\s+/, $v){
	if( $g eq '-ALL' ){
	    %g = ();
	    next;
	}
	if( $g =~ /^-(.*)/ ){
	    delete $g{$1};
	    next;
	}
	$g{$g} = 1;
    }

    $v = join ' ', keys %g;

    $v || undef;
}


# internally, only the extended version are used, in simple mode we calulate them from
# the specified simple set
# 
# user, staff, root =>
# page, about, getconf, flush, override, annotate, logfile, ntfylist, ntfyack, ntfyackall, ntfydetail

sub init_acls {
    my $me = shift;

    if( $me->{acl_mode} eq 'simple' ){

	$me->{acl_page}       = $me->{acl_user};
	
	$me->{acl_about}      =
	$me->{acl_getconf}    =
	$me->{acl_flush}      =
	$me->{acl_ntfyackall} = $me->{acl_root};
	
	$me->{acl_override}   =
	$me->{acl_annotate}   =
	$me->{acl_logfile}    =
	$me->{acl_ntfylist}   =
	$me->{acl_ntfyack}    =
	$me->{acl_ntfydetail} = $me->{acl_staff};
    }
}


# how in such short time, From eve to morn has the sun made his transit?
#      -- Dante, Divine Comedy

# something changed, determine current status
sub transition {
    my $me = shift;
    my $by = shift;
    my( $c, %cr, %co, $ca );
    my( $nr, $no );

    foreach $c (@{$me->{children}}){
	my $cc = $c;
	$cc = $c->aliaslookup() if $c->can('aliaslookup');
	next if $me->{nostatus};
	$cr{ $cc->{status} }++;
	$co{ $cc->{ovstatus} }++;
	$ca ++ if $cc->{alarm};
	
	my $cst = $cc->{sirentime} || 0;
	my $mst = $me->{sirentime} || 0;
	if( $cst > $mst ){
	    $me->{sirentime} = $cst;
	}
    }

    # QQQ - other states: degraded? concerned?
    if( $me->{gravity} eq "down" ){
	$nr = $cr{down} ? 'down' : 'up';
	$no = $co{down} ? 'down' : $co{override} ? 'override' : 'up';
    }else{
	$nr = $cr{up}   ? 'up'   : 'down';
	# do not propagate child-override in up gravity, else large agregates will likely never show down
	$no = $cr{up}   ? 'up'   : 'down';
    }

    if( $nr ne $me->{status} ){
	$me->{transtime} = $^T;
    }

    $me->{prevstatus}   = $me->{status};
    $me->{prevovstatus} = $me->{ovstatus};
    $me->{status}   = $nr;
    $me->{ovstatus} = $no;
    
    if( $ca && ($me->{status} eq "down" && $me->{siren}) ){
	$me->{alarm} = 1;
	# this may get cleared later if there is an override
    }else{
	$me->{alarm} = 0;
    }

    $me->transition2();
    $me->loggit( msg => ($by->{label_left} || $by->{label} || $by->{name}),
		 tag => 'TRANSITION',
		 lss => 1,
		 slp => 1 ) if $me->{prevovstatus} ne $me->{ovstatus};

    $me->transition_propagate();
}

sub transition2 {
    my $me = shift;

    if( $me->{override} ){
	# set override
	$me->{ovstatus} = 'override' unless $me->{status} eq 'up';

	# clear alarm, since we are in override
	$me->{alarm}    = 0;
	
	# clear override
	if( $me->{override} && $me->{status} eq 'up' && $me->{override}{mode} eq 'auto' ){
	    $me->override_remove();
	}
    }
    
}

# That labour on the bosom of this sphere, To propagate their states
#     -- Shakespeare, Timon of Athens

# propagate transition change to parents
sub transition_propagate {
    my $me = shift;
    my( $p );

    # send page
    if( $me->{prevstatus} && $me->{status} ne $me->{prevstatus}
	# do not notify if an ancestor is in override
	&& !$me->{anc_in_ov} ){
      Notify::new( $me )
    }

    # update hushed sirens
    if( $me->{prevovstatus} && ($me->{ovstatus} ne $me->{prevovstatus})
	&& ($me->{ovstatus} eq 'down') ){
	$me->{sirentime} = $^T;
    }
    
    # update stats
    $me->stats_transition();
    
    # webpage
    $me->{web}{transtime} = $^T;
    
    foreach $p ( @{$me->{parents}} ){
	$p->transition( $me );
    }
	
}

sub resolve_alii {
    my $me = shift;
    my $cf = shift;
    my( $c );

    if( $me->can('aliaslookup') ){
	$me->aliaslookup($cf);
    }
    foreach $c ( @{$me->{children}} ){
	$c->resolve_alii($cf);
    }
}

# The children of the porters: the children of Shallum, the children
# of Ater, the children of Talmon, the children of Akkub, the children
# of Hatita, the children of Shobai, in all an hundred thirty and
# nine.
#     -- Ezra 2:42
sub sort_children {
    my $me = shift;
    my( $c );

    if( $me->{sort} ){
	@{$me->{children}} = sort {
	    ($a->{label_left} || $a->{name}) cmp ($b->{label_left} || $b->{name})
	    } @{$me->{children}};
    }
    
    foreach $c ( @{$me->{children}} ){
	$c->sort_children();
    }
}

# need to make sure we recalc statuses as objects may have changed on disk
sub jiggle {
    my $me = shift;
    my( $c );

    foreach $c ( @{$me->{children}} ){
	$c->jiggle();
    }
}

################################################################
# class methods
################################################################

# has_attr gets called for every field for every object (or about 500 times per object)
# at a 1000 objects, this adds up to some significant time
# memoize...
my %hasattrcache = ();

sub has_attr {
    my $field = shift;
    my $doc   = shift;
    my $attr  = shift;
    my( $v );
    
    $v = $hasattrcache{ $doc->{package} }{$field}{$attr};
    return $v if defined $v;
    
    $v = grep {/$attr/} @{$doc->{fields}{$field}{attrs}};
    $hasattrcache{ $doc->{package} }{$field}{$attr} = $v;
    $v;
}

sub clearcache {
    %hasattrcache = ();
}

sub cmd_list {
    my $ctl = shift;
    my( $x );
    
    $ctl->ok();

    foreach $x (@all){
	$ctl->write( $x->unique() . "\n" );
    }
    $ctl->write( "\n" );
}

sub cmd_children {
    my $ctl = shift;
    my $param = shift;
    my( $x, $c );

    $x = $byname{ $param->{object} };
    if( $x ){
	$ctl->ok();
	foreach $c ( @{$x->{children}} ){
	    $ctl->write( $c->unique() . "\n" );
	}
	$ctl->write( "\n" );
    }else{
	$ctl->write( "BUMMER 404 Object Not Found\n\n" );
    }
}

sub cmd_about {
    my $ctl = shift;
    my $param = shift;
    my( $b, $x, $k, $v );
    
    $x = $byname{ $param->{object} };
    if( $x ){
	$ctl->ok();
	if( $x->{i_am_top} && (!$Conf::Config{_no_addtnl} || $param->{argusctl}) ){
	    # maybe dump some additional debugging info
	    $ctl->write( "__version_perl:\t" .
			 ( $^V ? join('.', unpack('c*', $^V)) : $] ) . "\n");
	    $ctl->write( "__version_argus:\t$::VERSION\n" );
	    $ctl->write( "__uptime:\t" . elapsed($^T - $::starttime) . " / " .
			 elapsed($^T - $::mainstart) . "\n" );
	    $ctl->write( "__osinfo:\t$::OSINFO\n" );
	    # ...
	}
	
	foreach $b ('', 'override', 'stats', 'web', 'notify'){
	    next unless !$b || $x->{$b};
	    foreach $k (sort ($b ? (keys %{$x->{$b}}) : (keys %$x))){
		$v = $b ? $x->{$b}{$k} : $x->{$k};
		next unless defined $v;
		if( ref($v) ){
		    $v = "#<REF>";
		}else{
		    # QQQ - or some other encoding scheme?
		    $v = ::xxx_encode( $v );
		}
		$ctl->write( $b ? "${b}::$k:\t$v\n" : "$k:\t$v\n" );
	    }
	}
	$x->about_more($ctl) if $x->can('about_more');
	$ctl->write( "\n" );
    }else{
	$ctl->write( "BUMMER 404 Object Not Found\n\n" );
    }
}

sub cmd_annotate {
    my $ctl = shift;
    my $param = shift;
    
    my $x = $byname{ $param->{object} };
    if( $x ){
	if( $param->{text} ){
	    $ctl->ok();
	    $x->{annotation} = $param->{text};
	    # splitting onto 2 lines looks nicer
	    $x->loggit( msg => "annotation: $param->{text}",
			tag => 'ANNOTATION',
			slp => 1 );
	    $x->loggit( msg => "annotation added by $param->{user} ",
			tag => 'ANNOTATION',
			slp => 1 );
	    $x->{web}{transtime} = $^T;
	}elsif( $x->{annotation} ){
	    # clear annotation
	    $ctl->ok();
	    delete $x->{annotation};
	    $x->{web}{transtime} = $^T;
	    $x->loggit( msg => "annotation cleared by $param->{user}",
			tag => 'ANNOTATION',
			slp => 1 );
	}else{
	    $ctl->write( "BUMMER 404 No annotation\n" );
	}
	$ctl->write( "\n" );
    }else{
	$ctl->write( "BUMMER 404 Object Not Found\n\n" );
    }
}

sub cmd_getconf {
    my $ctl = shift;
    my $param = shift;
    my( $c );
    
    my $x = $byname{ $param->{object} };
    if( $x ){
	$ctl->ok();
	$c = $x->gen_conf(0);
	$c =~ s/^/-/gm;
	$ctl->write( $c );
	$ctl->write( "\n" );
    }else{
	$ctl->write( "BUMMER 404 Object Not Found\n\n" );
    }
}

# NB: getparam and setparam are useful for debugging...
#     getparam is used by cgi to get various things...
sub cmd_getparam {
    my $ctl = shift;
    my $param = shift;
    my( $x, $z, $r, $p, @p );

    $x = $byname{ $param->{object} };
    if( $x ){
	$p = ::xxx_decode( $param->{param} );
	@p = split /::/, $p;
	$r = $x;
	foreach $z (@p){
	    if( $z =~ /^\d+$/ ){
		return $ctl->write( "BUMMER 404 Param Not Found\n\n" )
		    # this should be exists( $r->[$z] ) -- but in 5.00503
		    # exists requires a HASH, not an ARRAY...
		    unless ref($r) && defined( $r->[$z] );
		$r = $r->[$z];
	    }else{
		return $ctl->write( "BUMMER 404 Param Not Found\n\n" )
		    unless ref($r) && exists( $r->{$z} );
		$r = $r->{$z};
	    }
	}
	$ctl->ok();
	$ctl->write( "value: " . ::xxx_encode($r) . "\n\n" );
    }else{
	$ctl->write( "BUMMER 404 Object Not Found\n\n" );
    }
}

# setparam should be used with extreme caution, you can cause serious damage...  
# perhaps it should be removed/disabled in non-development versions?
# hmmm, well for now, here is a gun, don't shoot your foot.
sub cmd_setparam {
    my $ctl = shift;
    my $param = shift;
    my( $x, $z, $fz, $r, $rfr, $p, $v, @p );

    $x = $byname{ $param->{object} };
    $v = ::xxx_decode( $param->{value} );
    if( $x ){
	$p = ::xxx_decode( $param->{param} );
	@p = split /::/, $p;
	$r = $x;
	$fz = pop @p;
	foreach $z (@p){
	    if( $z =~ /^\d+$/ ){
		return $ctl->write( "BUMMER 404 Param Not Found\n\n" )
		    # see comment above, Re: exists() in 5.00503
		    unless ref($r) && defined( $r->[$z] );
		$r = $r->[$z];
	    }else{
		return $ctl->write( "BUMMER 404 Param Not Found\n\n" )
		    unless ref($r) && exists( $r->{$z} );
		$r = $r->{$z};
	    }
	}
	$ctl->ok();
	return $ctl->write( "BUMMER 404 Param Not Found\n\n" )
	    unless ref($r);
	$r->{$fz} = $v;
	$ctl->write( "\n" );
    }else{
	$ctl->write( "BUMMER 404 Object Not Found\n\n" );
    }
}

################################################################

Control::command_install( 'list',     \&cmd_list,
			  "return all objects" );
Control::command_install( 'children', \&cmd_children,
			  "return all children of specified object", "object" );
Control::command_install( 'about',    \&cmd_about,
			  "return internal data fields of specified object", "object" );
Control::command_install( 'annotate', \&cmd_annotate,
			  "add an annotation to an object--this will show on the webpppage",
			  "object user text" );
Control::command_install( 'getconf',  \&cmd_getconf,
			  "return the config data for an object", "object" );
Control::command_install( 'getparam', \&cmd_getparam ); # undocumented
Control::command_install( 'setparam', \&cmd_setparam ); # undocumented

Doc::register( $doc );

1;
