# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 23:24 (EST)
# Function: ping service test
#
# $Id: Ping.pm,v 1.13 2002/10/20 22:32:14 jaw Exp jaw $

package Ping;

@ISA = qw(Service BaseIO MonEl);

# this works by getting a bunch of objects at a time
# and fping'ing them all at once
# it plays some games with the scheduling to accomplish this

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [@ISA],
    methods => {},
    fields => {
      ping::addr => {},
      ping::rbuffer => {},
      ping::running => {},
      ping::data => {},
      ping::rtt => {
	  descr => 'round trip time of last ping test',
      },
    },
};

my %pending = ();

sub probe {
    my $name = shift;

    return ( $name =~ /^Ping/ ) ? \&config : undef;
}

sub config {
    my $me = shift;
    my $cf = shift;
    my( $ip );

    # $me->init_from_config( $doc, 'ping' ); # nothing to configure...
    $ip = ::resolve( $me->{srvc}{hostname} );
    return $cf->error( "cannot resolve host $me->{srvc}{hostname}" )
	unless $ip;
    
    $me->{ping}{addr} = ::inet_ntoa( $ip );
    $me->{uname} = "Ping_$me->{srvc}{hostname}";

    bless $me, 'Ping';
    $me;
}

sub start {
    my $me = shift;
    my( $fh, $x, $fn, $pid );

    if( $me->{ping}{running} ){
	$me->debug( "already pinging" );
	return;
    }
    $me->debug( "start ping" );
    
    $me->{fd} = $fh = BaseIO::anon_fh();

    # Jack shall pipe and Gill shall dance.
    #   -- George Wither, Poem on Christmas.
    
    # open pipes to fping, and fork,exec
    unless( pipe($fh, FPW) ){
	::loggit( "pipe1 fping failed: $!", 'PING', 0);
	$me->done();
	return;
    }
    unless( pipe(FPR, FPD) ){
	::loggit( "pipe2 fping failed: $!", 'PING', 0);
	$me->done();
	return;
    }
    
    unless( ($pid = fork) ){
        # child
        BaseIO::closeall();
	close STDOUT; open( STDOUT, ">&FPW" ); close FPW;
        close STDERR; open( STDERR, ">/dev/null" );
        close STDIN;  open( STDIN, "<&FPR" );  close FPR;
        close $fh;
        close FPD;
	# Execute their airy purposes.
        #   -- John Milton, Paradise Lost
        exec( "$::path_fping -r 5 -t 500 -e" ); # no, I didn't mean system() when I said exec()
	  # and yes, I am aware the following statement is unlikely to be reached.
        syswrite( STDERR, "bummer dude, couldn't fping, what's that about? $!\n" );
        _exit(-1);
    }

    # First Musician  Faith, we may put up our pipes, and be gone.
    #   -- Shakespeare, Romeo+Juliet
    close FPR;
    close FPW;

    if( !defined($pid) ){
	# fork failed
	close FPD;
	$me->done();
	return ::loggit( "fork fping failed: $!", 'PING', 0);
    }
    
    $me->{ping}{pid} = $pid;
    $me->baseio_init();
    $me->Service::start();

    $fn = fileno($fh);
    $me->wantread(1);
    $me->wantwrit(0);
    $me->settimeout(300);
    $me->{ping}{rbuffer} = '';

    # the profiler suggested that I find a better way to do this
    # in the name of speed, we'll poke around in what was formerly
    # BaseIO's private data...
    # ick!
    # but 0.0030s -> 0.0010s

    # find all of the things scheduled to be pinged at the same time
    # send them all to fping and do some bookkeeping so we can find
    # everything later
    foreach $x (@BaseIO::bytime){
	last if $x->{time} > $^T;
	my $o = $x->{obj};
	next unless defined $o->{ping};
	$o->{ping}{pid} = $pid;
	# save effort for BaseIO and self, clean up after ourself
	$x = undef;
	
	print FPD "$o->{ping}{addr}\n";
	push @{$pending{ "$fn/$o->{ping}{addr}" }}, $o;
	$o->{ping}{running} = 1;
	$o->{srvc}{state} = 'running';
	$o->debug( "also pinging" );
    }

    # also add current obj (it has already been removed from @bytime)
    print FPD "$me->{ping}{addr}\n";
    push @{$pending{ "$fn/$me->{ping}{addr}" }}, $me;
    $me->{ping}{running} = 1;
    $me->{srvc}{state} = 'running';
    $me->debug( "pinging" );
    
    close FPD;
}

sub timeout {
    my $me = shift;

    kill 9, $me->{ping}{pid} if $me->{ping}{pid};
    $me->finish();
}

sub readable {
    my $me = shift;
    my( $fh, $i, $l );

    $fh = $me->{fd};
    $i = sysread $fh, $l, 8192;
    if( $i ){
	$me->debug( "ping - read data ($l)" );
	$me->{ping}{rbuffer} .= $l;
    }else{
	$me->finish();
    }
}

# fping gives us lines like:
#   ip is alive (100 msec)
#   ip is unreachable
# for each line, find the correct ping object (from the table built in start)
# and mark them all as up/down
sub finish {
    my $me = shift;
    my( $fh, $fn, $ip, $x, $rtt, @l );

    $fh = $me->{fd};
    $fn = fileno($fh);

    @l = split /\n/, $me->{ping}{rbuffer};
    $me->{ping}{rbuffer} = '';
    $me->wantread(0);
    foreach (@l){
	($ip) = /^([^\s]+)\s/;
	($rtt) = /\((.*) msec/;
	foreach $x ( @{$pending{ "$fn/$ip" }} ){
	    next unless $x;
	    delete $pending{ "$fn/$ip" };
	    $x->{ping}{running} = undef;
	    $x->debug( "PING: $_" );
	    # keep the rtt, and full line available for debugging
	    $x->{ping}{rtt} = $rtt;
	    $x->{ping}{data} = $_;
	    if( /alive/ ){
		$x->isup();
	    }else{
		$x->isdown( $_ );
	    }
	}
    }
}

sub done {
    my $me = shift;

    $me->shutdown();
    $me->Service::done();
}

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->Service::about_more($ctl);
    foreach $k (sort keys %{$me->{ping}}){
	$v = $me->{ping}{$k};
	next unless defined $v;
	if( ref($v) ){
	    $v = "#<REF>";
	}else{
	    $v = ::xxx_encode( $v );
	}
	$ctl->write( "ping::$k:\t$v\n" );
    }
}

# sub webpage_more {
#     my $me = shift;
#     my $fh = shift;
#     my( $k, $v );
#     
#     foreach $k (qw(rtt)){
# 	  $v = $me->{ping}{$k};
# 	  print $fh "<TR><TD>$k</TD><TD>$v</TD></TR>\n" if defined($v);
#     }
# }

################################################################
# global config
################################################################
Doc::register( $doc );
push @Service::probes, \&probe;

1;
