# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 15:59 (EST)
# Function: testing of TCP services
#
# $Id: TCP.pm,v 1.17 2002/10/20 22:32:15 jaw Exp jaw $


package TCP;
@ISA = qw(Service BaseIO MonEl);

use Fcntl;
use Socket;

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [@ISA],
    methods => {},
    fields => {
      tcp::port => {
	  descr => 'TCP port to test',
	  attrs => ['config'],
      },
      tcp::addr => {
	  descr => 'encoded IP address',
      },
      tcp::send => {
	  descr => 'text to send once connected',
	  attrs => ['config'],
      },
      tcp::expect => {
	  descr => 'regex to expect',
	  attrs => ['config'],
      },
      tcp::url => {
	  descr => 'url to check',
	  attrs => ['config'],
      },
      tcp::browser => {
	  descr => 'browser spoofing for URL checks',
	  attrs => ['config', 'inherit'],
      },
	
      tcp::readhow => {
	  descr => 'how much data should be read from the server before checking expect',
	  attrs => ['config'],
	  vals  => ['banner', 'toeof', 'once'],
      },
      tcp::altsend => {},
      tcp::rbuffer => {},
      tcp::wbuffer => {},
      tcp::start   => {},
      tcp::rtt     => {},
    },

};

%config =
(

 SMTP   => {
     # send   => "MAIL\r\n",  # not RFC compliant, but quiets sendmails logs
     port   => 25,	expect => '^220',	readhow => "banner",
 },
 
 FTP    => {
     port   => 21,	expect => '^220',	readhow => "banner",
 },
	   
 POP    => {
     port   => 110,	expect => '^\+OK',	readhow => "banner",
 },
 
 NNTP   => {
     port   => 119,	expect => '^200',	readhow => "banner",
 },
	   
 HTTP   => {
     port   => 80,
     send   => "HEAD / HTTP/1.0\r\n\r\n",
     expect => "HTTP/",				readhow => "toeof",
 },

 # this is different than HTTP--it tests the content of a page
 # and will follow (to limited extent) http redirects
 # and can do browser spoofing (browser: Mozilla/4.0...)
 URL   => {
     expect => "HTTP/[0-9\.]+ 200",		readhow => 'toeof',
 },
 
 Telnet => {
     port => 23,
 },

 Gopher => {
     port => 70,
     send => "\r\n",
     expect  => "\.\r\n",			readhow => 'toeof',
 },
 
 SSH => {
     port => 22, expect => '^SSH',              readhow => 'banner',
 },

 IMAP => {
     port => 143, expect => '^* OK',            readhow => 'banner',
 },

 SSL => {
     port => 443,
 },
 HTTPS => {
     port => 443,
 },
 
 );


sub probe {
    my $name = shift;

    return ( $name =~ /^TCP/ ) ? \&config : undef;
}

sub config {
    my $me = shift;
    my $cf = shift;
    my( $name );

    $name = $me->{name};
    $name =~ s/^TCP\/?//;

    $me->init_from_config( $doc, 'tcp' );
    
    $me->{tcp}{port}        ||= $config{$name}{port};
    $me->{tcp}{send}        ||= $config{$name}{send};
    $me->{tcp}{expect}      ||= $config{$name}{expect};
    $me->{tcp}{readhow}     ||= $config{$name}{readhow};
    
    $me->{tcp}{hostname}      = $me->{srvc}{hostname};
    
    $me->{label_right_maybe} ||= $name;

    if( $name eq "URL" ){
	my( $host, $port, $file ) = $me->{tcp}{url} =~ m|http://([^:/]+)(?::([^/]+))?(/.*)?|;
	$me->{tcp}{port} = $port ||= 80;
	$file ||= "/";
	$me->{srvc}{hostname} = $me->{tcp}{hostname} = $host;
	$me->{uname} = "URL_$host:$port$file";
	unless( $host ){
	    $me->{transient} = 1;
	    return $cf->error( "invalid URL" );
	}
	$me->{tcp}{send}   = "GET $file HTTP/1.0\r\nHost: $host:$port\r\n";
	if( $me->{tcp}{browser} ){
	    $me->{tcp}{send}   .= "User-Agent: $me->{tcp}{browser}\r\n";
	}
	$me->{tcp}{send} .= "\r\n";
    }else{
	if( $name ){
	    $me->{uname} = "${name}_$me->{srvc}{hostname}";
	}else{
	    $me->{uname} = "TCP_$me->{tcp}{port}_$me->{srvc}{hostname}";
	}
    }
    
    $me->{tcp}{addr} = ::resolve( $me->{srvc}{hostname} );
    unless( $me->{tcp}{addr} ){
	$me->{transient} = 1; # to prevent save() when it gets recycle()d 
	return $cf->error( "cannot resolve host $me->{srvc}{hostname}" );
    }

    unless( $me->{tcp}{port} ){
	$me->{transient} = 1;
	return $cf->error( "Incomplete specification or unknown protocol for Service $name" );
    }

    bless $me, 'TCP';

    $me;
}

sub start {
    my $me = shift;
    my( $i, $fh, $to );

    $me->{fd} = $fh = BaseIO::anon_fh();

    $i = socket($fh, PF_INET, SOCK_STREAM, getprotobyname('tcp'));
    $me->baseio_init();
    $me->Service::start();
    
    $SIG{ALRM} = sub {};
    alarm(5);
    # in some cases this may block, even though we set non-blocking
    connect( $fh, sockaddr_in($me->{tcp}{port}, $me->{tcp}{addr}) );
    # if the connect fails, we'll pick up the error later
    alarm(0);
    
    $me->{srvc}{state} = "connecting";
    $me->wantread(0);
    $me->wantwrit(1);
    $me->settimeout( $me->{srvc}{timeout} );
    $me->{tcp}{rbuffer} = '';
    $me->{tcp}{wbuffer} = '';
    $me->{tcp}{start} = $^T;
    
    $me->debug( "TCP Start: connecting - $me->{tcp}{port}, ".
		"$me->{srvc}{hostname}, try $me->{srvc}{tries}" );
    
}

sub timeout {
    my $me = shift;

    $me->isdown( "TCP timeout" );
}

sub http_redirect {
    my $me = shift;
    my( $url, $file, $loc, @lines );

    # NB: this code path never calls done
    $me->shutdown();
    
    @lines = split( /\n/, $me->{tcp}{rbuffer} );
    ($loc) = grep /^Location:/, @lines;
    $loc =~ tr/\r//d;
    ($url) = $loc =~ /^Location:\s+(.*)/;
    ($file) = $url =~ /http:\/\/[^\/]*(.*)/;
    # NB: cannot redirect to another host
    
    $me->{tcp}{altsend}   = "GET $file HTTP/1.0\r\nHost: $me->{tcp}{hostname}:$me->{tcp}{port}\r\n";
    if( $me->{tcp}{browser} ){
	$me->{tcp}{altsend}   .= "User-Agent: $me->{tcp}{browser}\r\n";
    }
    $me->{tcp}{altsend} .= "\r\n";
    $me->debug( "HTTP Redirect -> $url" );
    # start over without rescheduling
    $me->start();
}

sub writable {
    my $me = shift;

    if( $me->{srvc}{state} eq "connecting" ){
	$me->debug( "TCP - connected" );
	if( $me->{tcp}{send} ){
	    if( $me->{tcp}{altsend} ){
		$me->{tcp}{wbuffer} = $me->{tcp}{altsend};
		$me->{tcp}{altsend} = undef;
	    }else{
		$me->{tcp}{wbuffer} = $me->{tcp}{send};
	    }
	}
	$me->{srvc}{state} = "sending";
	$me->settimeout( $me->{srvc}{timeout} );
    }
    
    if( $me->{tcp}{wbuffer} ){
	my( $b, $i, $l, $fh );

	$fh = $me->{fd};
	$b = $me->{tcp}{wbuffer};
	$l = length($b);
	# And he asked for a writing table, and wrote
	#   -- Luke 1:63
	$i = syswrite $fh, $b, $l;
	if( $i ){
	    $me->debug( "TCP - wrote $i bytes of $l" );
	}else{
	    return $me->isdown( "TCP write failed: $!" );
	}
	$b = substr( $b, $i, $l );
	$me->{tcp}{wbuffer} = $b;
    }
    
    if( !$me->{tcp}{wbuffer} ){
	if( $me->{tcp}{expect} ){
	    $me->{srvc}{state} = "expecting";
	    $me->wantread(1);
	    $me->wantwrit(0);
	    $me->settimeout( $me->{srvc}{timeout} );
	}else{
	    # success is a connected socket
	    $me->isup();
	}
    }
}

sub readable {
    my $me = shift;
    my( $fh, $i, $l, $testp );

    $fh = $me->{fd};
    $i = sysread $fh, $l, 8192;
    # And this is the writing that was written: MENE, MENE, TEKEL, UPHARSIN.
    #   -- daniel 5:25
    if( $i ){
	$me->debug( "TCP - read data" );
	$me->{tcp}{rbuffer} .= $l;
    }
    elsif( defined($i) ){
        # And it came to pass, when Moses had made an end of writing ...
	#   -- dueteronomy 31:24
	# $i is 0 -> eof
	$testp = 1;
    }
    else{
	# $i is undef -> error
	return $me->isdown( "TCP read failed: $!" );
    }

    if( $me->{tcp}{readhow} eq "banner" && $me->{tcp}{rbuffer} =~ /\n/ ){
	# My ears have not yet drunk a hundred words
	# Of that tongue's utterance, yet I know the sound:
	#   -- Shakespeare, Romeo+Juliet
	$testp = 1;
    }
    if( $me->{tcp}{readhow} eq "once" ){
	$testp = 1;
    }

    if( $testp ){
	my( $e );
	# There is a written scroll! I'll read the writing.
	#   -- Shakespeare, Merchant of Venice
	$e = $me->{tcp}{expect};
	if( $me->{tcp}{rbuffer} =~ /$e/ ){
	    # and the writing was the writing of God
	    #   -- exodus 32:16
	    return $me->isup();
	}else{
	    if( $me->{name} eq "TCP/URL" && $me->{tcp}{rbuffer} =~ /HTTP\/1\.\d+\s+30[12]/ ){
		# once or twice she had peeped into the book her sister was reading, but
		# it had no pictures or conversations in it,
		#   -- Alice in Wonderland
		# try to handle redirect
		return $me->http_redirect();
	    }else{
		# Alas, Malvolio, this is not my writing
		#   -- Shakespeare, 12th Night
		return $me->isdown( "TCP - did not get expected data" );
	    }
	}
    }else{
	# kibbles and bits, kibbles and bits, give me more kibbles and bits
	#   -- Dog Food Commercial
	$me->wantread(1);
	$me->wantwrit(0);
    }
}



################################################################
# and also object methods
################################################################

sub done {
    my $me = shift;

    $me->{tcp}{rtt} = $^T - $me->{tcp}{start};
    $me->shutdown();
    $me->Service::done();
}

# sub isdown {
#     my $me = shift;
#     my $reason = shift;
# 
#     $me->Service::isdown($reason);
# }
# 
# sub isup {
#     my $me = shift;
# 
#     $me->Service::isup();
# }

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->Service::about_more($ctl);
    foreach $k (sort keys %{$me->{tcp}}){
	$v = $me->{tcp}{$k};
	next unless defined $v;
	if( ref($v) ){
	    $v = "#<REF>";
	}else{
	    $v = ::xxx_encode( $v );
	}
	$ctl->write( "tcp::$k:\t$v\n" );
    }
}

sub webpage_more {
    my $me = shift;
    my $fh = shift;
    my( $k, $v );
    
    foreach $k (qw(hostname port url)){
	$v = $me->{tcp}{$k};
	next unless defined $v;
	# I see a lot of long URLs...
	if( $k eq 'url' ){
	    my $c = $v;
	    $c = substr($c, 0, 70) . "..."
		if( length($c) > 70 );
	    $v = qq{<A HREF="$v">$c</A>};
	}
	print $fh "<TR><TD>$k</TD><TD>$v</TD></TR>\n";
    }
}

    
################################################################
# global config
################################################################
Doc::register( $doc );
push @Service::probes, \&probe;



1;
