# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 18:25 (EST)
# Function: testing of UDP services
#
# $Id: UDP.pm,v 1.21 2002/10/20 22:32:15 jaw Exp jaw $


package UDP;
@ISA = qw(Service BaseIO MonEl);

use Fcntl;
use Socket;
use SNMP;

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [@ISA],
    methods => {},
    fields => {
      udp::port => {
	  descr => 'UDP port to test',
	  attrs => ['config'],
      },
      udp::addr => {
	  descr => 'encoded IP address',
      },
      udp::send => {
	  descr => 'text to send once connected',
	  attrs => ['config'],
      },
      udp::expect => {
	  descr => 'regex to expect',
	  attrs => ['config'],
      },
      udp::rbuffer => {},
      udp::wbuffer => {},
	
      snmp::community => {
	  descr => 'SNMP community',
	  attrs => ['config', 'inherit'],
      },
      snmp::snmpcalc => {
	  attrs => ['config'],
      },
      snmp::snmpcalcdata => {},
      snmp::currvalue => {
	  descr => 'most recent value returned by query',
      },
      snmp::oid => {
	  descr => 'SNMP OID to quesry',
	  attrs => ['config'],
      },
	      
      udp::decodesnmp => {},

	# test value fields, 
      snmp::minvalue => {
	  descr => 'SNMP test value - fail test if value falls below this',
	  attrs => ['config'],
      },
      snmp::maxvalue => {
	  descr => 'SNMP test value - fail test if value rises above this',
	  attrs => ['config'],
      },
      snmp::eqvalue => {
	  descr => 'SNMP test value - fail test if value does not equal this',
	  attrs => ['config'],
      },
      snmp::nevalue => {
	  descr => 'SNMP test value - fail test if value equals this',
	  attrs => ['config'],
      },

      dns::zone => {
	  descr => 'DNS zone to test',
	  attrs => ['config'],
      },
	
    },
};

%config =
(
 # not supported by all radius servers
 RADIUS => {
     port => 1645, send => pack( "CCSx20", 12, $$, 20 ),
     timeout => 10,
 },
 
 DNS => { # RFC 1035, op=status
     port => 53,   send => pack("n CC nnnn", $$, 0x10,0, 0,0,0,0),
     timeout => 12,
 },

 DNSQ => { # djb does not respond to status queries
     port => 53, timeout => 12,
     send => pack( "n CC nnnn  Cnn", $$, 1,0, 1,0,0,0, 0,255,1),
     # op=query, recurse, 1 query, . ANY IN
 },
 
 # test server authority for zone
 Domain => {
     port => 53, timeout => 12,
 },
 
 SNMP => {
     port => 161,  timeout => 5, 
 },

 NTP => { # RFC 2030
     port => 123,  timeout => 10,
     send => pack("CCCC x64", 0x23,0,0,0),
 },

 # NFS = RFC 1094; RPC = RFC 1057; XDR = RFC 1014
 # NFSv2 - NFSPROC_NULL
 NFS => {
     port => 2049, timeout => 10,
     send => pack( "NN NN NN x16", $$, 0,  2, 100003, 2, 0),
     # xid, type, rpcver, prog, ver, func, cred(flavor, len, null), verf(flavor, len, null)
 },

 NFSv3 => {
     port => 2049, timeout => 10,
     send => pack( "NN NN NN x16", $$, 0,  2, 100003, 3, 0),
 },

 );


sub probe {
    my $name = shift;

    return ( $name =~ /^UDP/ ) ? \&config : undef;
}

sub config {
    my $me = shift;
    my $cf = shift;
    my( $name, $base );
    
    bless $me, 'UDP';

    $name = $me->{name};
    $name =~ s/^UDP\/?//;
    $base = $name;
    $base =~ s/\/.*//;

    $me->{srvc}{timeout}     = $config{$base}{timeout};
    
    $me->init_from_config( $doc, 'udp' );

    $me->{udp}{port}        ||= $config{$base}{port};
    $me->{udp}{send}        ||= $config{$base}{send};
    $me->{udp}{expect}      ||= $config{$base}{expect};
    
    $me->{udp}{hostname}    = $me->{srvc}{hostname};
    
    if( $name ){
	$me->{uname} = "${name}_$me->{srvc}{hostname}";
    }else{
	$me->{uname} = "UDP_$me->{udp}{port}_$me->{srvc}{hostname}";
    }

    $me->{udp}{addr} = ::resolve( $me->{srvc}{hostname} );
    unless( $me->{udp}{addr} ){
	$me->{transient} = 1;
	return $cf->error( "cannot resolve host $me->{srvc}{hostname}" );
    }

    if( $name =~ /^SNMP/ ){
	$me->init_from_config( $doc, 'snmp' );
	if( $name =~ /SNMP\/(.*)/ ){
	    $me->{snmp}{oid} ||= $1;
	    $name = $1;
	}else{
	    $me->{snmp}{oid} ||= $config{$name}{oid}
	    	if $config{$name};
	    $name = $me->{snmp}{oid};
	}

	$me->{uname} = "SNMP_$me->{snmp}{oid}_$me->{srvc}{hostname}";
	return undef unless $me->snmp_init($cf);
    }
    if( $name =~ /^Domain/i ){
	# test authoratativeness of nameserver for domain
	my($q, @q);
	$me->init_from_config( $doc, 'dns' );
	if( $name =~ /\/(.*)/ ){
	    $me->{dns}{zone} ||= $1;
	    $name = $1;
	}
	return $cf->error("zone for domain query?") unless $me->{dns}{zone};

	$q = pack( "n CC nnnn", 0, 0,0, 1,0,0,0);  # query, 1 question
	@q = split /\./, $me->{dns}{zone};
	foreach my $l (@q){
	    $q .= chr(length($l)) . $l;
	}
	$q .= "\0";
	$q .= "\0\6\0\1";   # SOA IN

	$me->{udp}{send} = $q;
	$me->{debug} = 1;
    }
    
    $me->{label_right_maybe} ||= $name;

    unless( $me->{udp}{port} ){
	$me->{transient} = 1;
	return $cf->error( "Incomplete specification or unknown protocol for Service $name" );
    }

    $me;
}

sub start {
    my $me = shift;
    my( $fh, $i );

    $me->{fd} = $fh = BaseIO::anon_fh();
    $i = socket($fh, PF_INET, SOCK_DGRAM, getprotobyname('udp'));

    $me->baseio_init();
    $me->Service::start();
    
    $me->debug( "UDP start" );
    $i = bind($fh, sockaddr_in(0, INADDR_ANY) );
    return $me->isdown( "UDP bind failed: $!" )
	unless $i;

    # And they made the plate of the holy crown of pure gold, and wrote upon it a writing
    #   -- exodus 39:30
    $i = send( $fh, $me->{udp}{send}, 0, sockaddr_in($me->{udp}{port}, $me->{udp}{addr}) );
    return $me->isdown( "UDP send failed: $!" )
	unless $i;

    $me->debug( "UDP sent data" );

    $me->wantread(1);
    $me->wantwrit(0);
    $me->settimeout( $me->{srvc}{timeout} );
    $me->{srvc}{state} = "waiting";

}

sub readable {
    my $me = shift;
    my( $fh, $i, $l, $e );

    $fh = $me->{fd};
    $i = recv($fh, $l, 8192, 0);
    return $me->isdown( "UDP recv failed: $!" )
	unless defined($i);

    $me->debug( "UDP recv data" );
    $me->{udp}{rbuffer} = $l;		# for debugging
    
    # yet I will read the writing unto the king, and make known to him the interpretation
    #   -- daniel 5:17

    # this should probably be done with a subclass, or at least a sub ref
    if( $me->{udp}{decodesnmp} ){
	# decode snmp
	return $me->snmp_recv( $l );
    }elsif( $me->{dns}{zone} ){
	my $a = vec($l, 2, 8);
	my $b = vec($l, 3, 8);

	if( $b & 0xF ){
	    $me->isdown( 'DNS Error (RCODE=' . ($b&0xF) . ')' );
	}
	elsif( $a & 4 ){
	    $me->isup();
	}else{
	    # He looked at the Gryphon as if he thought it had some kind of authority over Alice.
	    #   -- Alice in Wonderland
	    $me->isdown( 'DNS Error - Server Not Authoritative for Zone' );
	}
	
    }else{
	$e = $me->{udp}{expect};
	if( $e ){
	    if( $l =~ /$e/ ){
		$me->isup();
	    }else{
		# O, when she had writ it and was reading it over, she
                # found Benedick and Beatrice between the sheet?
		#   -- Shakespeare, Much Ado About Nothing
		$me->isdown( "UDP - did not get expected data" );
	    }
	}else{
	    $me->isup();
	}
    }
}

sub writable {
    my $me = shift;
    # NOT REACHED
}

sub timeout {
    my $me = shift;

    $me->isdown( "UDP timeout" );
}

################################################################
# and also object methods
################################################################

sub done {
    my $me = shift;

    $me->shutdown();
    $me->Service::done();
}

# sub isdown {
#     my $me = shift;
#     my $reason = shift;
# 
#     $me->Service::isdown($reason);
# }
# 
# sub isup {
#     my $me = shift;
# 
#     $me->Service::isup();
# }

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->Service::about_more($ctl);

    foreach $b ('udp', 'snmp', 'dns'){
	next unless $me->{$b};
	foreach $k (sort keys %{$me->{$b}}){
	    $v = $me->{$b}{$k};
	    next unless defined $v;
	    if( ref($v) ){
		$v = "#<REF>";
	    }else{
		# QQQ - or some other encoding scheme?
		$v = ::xxx_encode( $v );
	    }
	    $ctl->write( "${b}::$k:\t$v\n" );
	}
    }
}

sub webpage_more {
    my $me = shift;
    my $fh = shift;
    my( $k, $v );
    
    foreach $k (qw(hostname port)){
	$v = $me->{udp}{$k};
	print $fh "<TR><TD>$k</TD><TD>$v</TD></TR>\n" if defined($v);
    }
    foreach $k (qw(oid currvalue)){
	$v = $me->{snmp}{$k};
	print $fh "<TR><TD>SNMP $k</TD><TD>$v</TD></TR>\n" if defined($v);
    }
    foreach $k (qw(zone)){
	$v = $me->{dns}{$k};
	print $fh "<TR><TD>DNS $k</TD><TD>$v</TD></TR>\n" if defined($v);
    }

}

################################################################
# global config
################################################################
Doc::register( $doc );
push @Service::probes, \&probe;

1;

